/*------------------------------------------------------------------------------
* Alpha Shooter
* A 3D OpenGL First-Person Shooter game
* Copyright (C) 2005, 2006, 2007 Nicola Cocchiaro <ncocchiaro@users.sf.net>
* Alpha Shooter can be found at http://www.sf.net/projects/alphashooter.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 or, (at your option) any later
* version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*-------------------------------------------------------------------------------
*/

// camera.c
// camera manipulation functions

#include "camera.h"
#include "fps.h"
#include "quatlib.h"
#include <stdio.h>

/* Prints camera position vectors */
void print_camera()
{
	fprintf(stdout, "\n");
	print_vector(cam->up, "up");
	print_vector(cam->pos, "pos");
	print_vector(cam->view, "view");
	print_vector(cam->right, "right");
	fprintf(stdout, "\n");
}

/* Builds a single rotation matrix from the orientation quaternion, to take 
 * out vectors from that will be passed to gluLookAt(). Also updates whatever 
 * rotation procedure needs to be updated (eg. auto-leveling) */
void update_rotation(GLdouble r[16])
{
	if (cam->leveling)
		auto_level(cam);
	
	q_convert_m(q_conjugate(cam->orientation), r);
}

/* Changes x and z values if necessary (hack) */
void restrict_zx(vector *v, int *res, GLdouble x)
{
	if ((cam->old_pos.z >= 22.0) || (cam->old_pos.z <= 12.0))
	{
		if (cam->old_pos.z >= 22.0)
		{
			v->z = 22.0;
			*res = *res + 1;
		}
		else if (cam->old_pos.z <= 12.0)
		{
			v->z = 12.0;
			*res = *res + 1;
		}
	}
	else
	{
		v->x = x;
		*res = *res + 1;
	}
}

/* Determines whether camera is going out of boundaries or not */
int out_of_bounds(vector *v)
{
	int res = 0;
	
	// y
	if (v->y < 3.0)
	{
		v->y = 3.0;
		res++;
	}
	else if (v->y > 13.0)
	{
		v->y = 13.0;
		res++;
	}
	
	// x
	if (v->x > 22.0)
	{
		v->x = 22.0;
		res++;
	}
	else if (v->x < -22.0)
	{
		v->x = -22.0;
		res++;
	}

	// z
	if (v->z > 29.0)
	{
		v->z = 29.0;
		res++;
	}
	else if (v->z < -29.0)
	{
		v->z = -29.0;
		res++;
	}
	// monitors
	else if ((v->z <= 22.0) && (v->z >= 12.0))
	{
		if ((v->y > 0.0) && (v->y <= 4.0))
		{
			if (v->x <= -19.0)
				restrict_zx(v, &res, -19.0);
			else if (v->x >= 19.0)
				restrict_zx(v, &res, 19.0);
		}
		else if ((v->y > 4.0) && (v->y <= 5.0))
		{
			if (v->x <= -19.5)
				restrict_zx(v, &res, -19.5);
			else if (v->x >= 19.5)
				restrict_zx(v, &res, 19.5);
		}
		else if ((v->y > 5.0) && (v->y <= 6.0))
		{
			if (v->x <= -20.5)
				restrict_zx(v, &res, -20.5);
			else if (v->x >= 20.5)
				restrict_zx(v, &res, 20.5);
		}
		else if ((v->y > 6.0) && (v->y <= 7.0))
		{
			if (v->x <= -21.5)
				restrict_zx(v, &res, -21.5);
			else if (v->x >= 21.5)
				restrict_zx(v, &res, 21.5);
		}
	}
	// fence
	else if ((v->z <= -8.15) && (v->z >= -31.0))
	{
		if ((v->y > 0.0) && (v->y <= 5.0))
		{
			if ((v->x <= -8.85) && (v->x >= -9.85))
			{
				v->x = -8.85;
				res++;
			}
			else if ((v->x <= -9.85) && (v->x >= -10.85))
			{
				v->x = -10.85;
				res++;
			}
			else if ((v->x <= -8.85) && (v->x >= -21.0))
			{
				if ((v->z <= -8.15) && (v->z >= -9.15))
				{
					v->z = -8.15;
					res++;
				}
				else if ((v->z <= -9.15) && (v->z >= -10.15))
				{
					v->z = -10.15;
					res++;
				}
			}
		}
	}
	return res;
}

/* Updates translation matrix according to shift and a rotation matrix
 * (relative motion) */
void update_translation(GLdouble r[16], GLdouble t[16])
{
	GLdouble x, y, z;
	vector v;

	x = cam->to_shift_x;
	y = cam->to_shift_y;
	z = cam->to_shift_z;
	
	t[0] = 1.0;
	t[1] = 0.0;
	t[2] = 0.0;
	t[3] = x*r[0] + y*r[4] + z*r[8] + cam->old_pos.x;
	t[4] = 0.0;
	t[5] = 1.0;
	t[6] = 0.0;
	if (state->jetpack)
		t[7] = x*r[1] + y*r[5] + z*r[9] + cam->old_pos.y;
	else
	{
		// freefall
		if (cam->old_pos.y > 3.0)
		{
			if (fps->fps > 1.0)
				t[7] = cam->old_pos.y - (0.05*100.0/fps->fps);
			else
				t[7] = cam->old_pos.y - 0.05;
		}
		else	
			t[7] = cam->old_pos.y;
	}
	t[8] = 0.0;
	t[9] = 0.0;
	t[10] = 1.0;
	t[11] = x*r[2] + y*r[6] + z*r[10] + cam->old_pos.z;
	t[12] = 0.0;
	t[13] = 0.0;
	t[14] = 0.0;
	t[15] = 1.0;

	cam->to_shift_x = 0.0;
	cam->to_shift_y = 0.0;
	cam->to_shift_z = 0.0;

	// check if new position is correct
	v.x = t[3];
	v.y = t[7];
	v.z = t[11];
	v.w = 1.0;
	if (out_of_bounds(&v))
	{
		t[3] = v.x;
		t[7] = v.y;
		t[11] = v.z;
	}
}

/* Recalculates camera vectors */
void update_camera()
{
	GLdouble r[16], t[16];
	
	memset(r, 0.0, 16*sizeof(GLdouble));

	update_rotation(r);
	update_translation(r, t);
	
	cam->old_pos.x = t[3];
	cam->old_pos.y = t[7];
	cam->old_pos.z = t[11];
	cam->old_pos.w = 1.0;
	matrix_multiply(t, r, cam->tot);
	orthonormalize_matrix(cam->tot);
	cam->pos.x = cam->tot[3];
	cam->pos.y = cam->tot[7];
	cam->pos.z = cam->tot[11];
	cam->pos.w = 1.0;
	cam->old_pos.x = cam->pos.x;
	cam->old_pos.y = cam->pos.y;
	cam->old_pos.z = cam->pos.z;
	cam->old_pos.w = 1.0;
	cam->view.x = cam->pos.x + cam->tot[8];
	cam->view.y = cam->pos.y + cam->tot[9];
	cam->view.z = cam->pos.z + cam->tot[10];
	cam->view.w = 1.0;
	cam->up.x = cam->tot[4];
	cam->up.y = cam->tot[5];
	cam->up.z = cam->tot[6];
	cam->up.w = 1.0;
	cam->right.x = cam->tot[0];
	cam->right.y = cam->tot[1];
	cam->right.z = cam->tot[2];
	cam->right.w = 1.0;
}

/* Resets camera position */
void reset_camera()
{
	cam->tot[0] = 1.0;
	cam->tot[1] = 3.0;
	cam->tot[2] = 0.0;
	cam->tot[4] = 0.0;
	cam->tot[5] = 4.0;
	cam->tot[6] = 0.0;
	cam->tot[8] = 0.0;
	cam->tot[9] = 3.0;
	cam->tot[10] = 1.0;
	cam->tot[12] = 0.0;
	cam->tot[13] = 0.0;
	cam->tot[14] = 0.0;
	cam->tot[15] = 1.0;

	cam->orientation.v.x = 0.0;
	cam->orientation.v.y = 0.0;
	cam->orientation.v.z = 0.0;
	cam->orientation.v.w = 1.0;
	cam->orientation.s = 1.0;
	
	cam->pos.x = 0.0;
	cam->pos.y = 3.0;
	cam->pos.z = 0.0;
	cam->pos.w = 1.0;
	cam->view.x = 0.0;
	cam->view.y = 3.0;
	cam->view.z = 1.0;
	cam->view.w = 1.0;
	cam->up.x = 0.0;
	cam->up.y = 4.0;
	cam->up.z = 0.0;
	cam->up.w = 1.0;
	cam->right.x = 1.0;
	cam->right.y = 3.0;
	cam->right.z = 0.0;
	cam->right.w = 1.0;

	cam->leveling = 0;
}

/* Find a rotation quaternion around the current X axis and multiply it to the 
 * current orientation */
void rotate_around_X(GLdouble angle, camera_t *cam)
{
	quaternion tmp;
	vector x_axis = {1.0, 0.0, 0.0, 1.0};
	quaternion res;

	apply_q2v(cam->orientation, &x_axis);
	tmp = aa_convert_q(x_axis, angle);
	res = q_multiply(cam->orientation, tmp);
	res = q_normalize(res);
	memcpy(&cam->orientation, &res, sizeof(quaternion));
}

/* Find a rotation quaternion around the current Y axis and multiply it to the 
 * current orientation */
void rotate_around_Y(GLdouble angle, camera_t *cam)
{
	quaternion tmp;
	vector y_axis = {0.0, 1.0, 0.0, 1.0};
	quaternion res;

	if (state->jetpack)
		apply_q2v(cam->orientation, &y_axis);
	tmp = aa_convert_q(y_axis, angle);
	res = q_multiply(cam->orientation, tmp);
	res = q_normalize(res);
	memcpy(&cam->orientation, &res, sizeof(quaternion));
}

/* Find a rotation quaternion around the current Z axis and multiply it to the 
 * current orientation */
void rotate_around_Z(GLdouble angle, camera_t *cam)
{
	quaternion tmp;
	vector z_axis = {0.0, 0.0, 1.0, 1.0};
	quaternion res;

	apply_q2v(cam->orientation, &z_axis);
	tmp = aa_convert_q(z_axis, angle);
	res = q_multiply(cam->orientation, tmp);
	res = q_normalize(res);
	memcpy(&cam->orientation, &res, sizeof(quaternion));
}

/* Levels the camera according to world up and right (strafe) vectors */
void auto_level(camera_t *cam)
{
	vector y_axis = { 0.0, 1.0, 0.0, 1.0 };
		
	if (get_parallel(cam->up, y_axis, cam) == 0)
		cam->leveling = 0;
}

// Rotates the camera so that the passed local camera axis is parallel to the 
// target vector
int get_parallel(vector target, vector local_axis, camera_t *cam)
{
	vector rot_axis, tmp_axis, t_local;
	GLdouble step = fps->rot;
	GLdouble angle_vec = 0.0, angleT = 0.0;
	quaternion tmp, res;

	// copy vector locally and find vector from the origin
	copy_vector(local_axis, &tmp_axis);
	//tmp_axis.x -= cam->pos.x;
	//tmp_axis.y -= cam->pos.y;
	//tmp_axis.z -= cam->pos.z;
	
	// find angle between up vector and target axis
	angle_vec = 180.0*angle_between_vectors(target, tmp_axis)/M_PI;
	
	if ((angle_vec > 0.005) && (angle_vec < 179.995))
	{
		// find axis to rotate around to get parallel to target axis
		rot_axis = cross_product(target, tmp_axis);
		
		// going towards angle_vec == 180.0
		if (angle_vec > 90.0)
		{
			// clamp rotation to target position
			if (angle_vec + step > 180.0)
				step = 180.0 - angle_vec;
		}
		// going towards angle_vec == 0.0
		else
		{
			// clamp rotation to target position
			if (angle_vec - step < 0.0)
				step = angle_vec;
		}

		// check if rotating a positive angle takes us closer...
		// try a rotation and recalculate angle
		// FIXME: this could probably be done a lot better
		t_local = normalize(find_rotated_coords(tmp_axis, step, rot_axis.x, rot_axis.y, rot_axis.z));
		angleT = 180.0*angle_between_vectors(t_local, target)/M_PI;
		
		// if it doesn't take us closer to the target,
		// rotate opposite way instead
		if ((angle_vec < 90.0) && (angleT > angle_vec))
			step = -step;
		
		tmp = aa_convert_q(rot_axis, step);
		
		res = q_multiply(cam->orientation, tmp);
		res = q_normalize(res);
		memcpy(&cam->orientation, &res, sizeof(quaternion));

		return 1;
	}
	else
		return 0;
}
