/*------------------------------------------------------------------------------
* Alpha Shooter
* A 3D OpenGL First-Person Shooter game
* Copyright (C) 2005, 2006, 2007 Nicola Cocchiaro <ncocchiaro@users.sf.net>
* Alpha Shooter can be found at http://www.sf.net/projects/alphashooter.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 or, (at your option) any later
* version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*-------------------------------------------------------------------------------
*/

// quatlib.c
// a simple quaternion library

#include <stdio.h>
#include <stdlib.h>
#include "quatlib.h"
#include "lib.h"

/* returns the conjugate of a quaternion */
quaternion q_conjugate(quaternion q)
{
	quaternion c;

	c.s = q.s;
	c.v.x = -q.v.x;
	c.v.y = -q.v.y;
	c.v.z = -q.v.z;
	c.v.w = q.v.w;

	return(c);
}

/* calculates the magnitude of a quaternion */
GLdouble q_magnitude(quaternion q)
{
	return(sqrt(q.s*q.s + q.v.x*q.v.x + q.v.y*q.v.y + q.v.z*q.v.z));
}

/* normalizes a quaternion */
quaternion q_normalize(quaternion q)
{
	GLdouble m = q_magnitude(q);

	if (m != 0.0)
	{
		q.s = q.s/m;
		q.v.x = q.v.x/m;
		q.v.y = q.v.y/m;
		q.v.z = q.v.z/m;
		q.v.w = 1.0;
	}

	return(q);
}

/* returns the inverse of a quaternion (equivalent to the conjugate if the 
 * quaternion is normalized) */
quaternion q_inverse(quaternion q)
{
	quaternion i;

	i = q_normalize(q);
	i = q_conjugate(i);

	return(i);
}

/* returns the quaternion resulting from the multiplication of two 
 * quaternions */
quaternion q_multiply(quaternion q, quaternion r)
{
	quaternion s;

	s.s = q.s*r.s - q.v.x*r.v.x - q.v.y*r.v.y - q.v.z*r.v.z;
	s.v.x = q.s*r.v.x + q.v.x*r.s + q.v.y*r.v.z - q.v.z*r.v.y;
	s.v.y = q.s*r.v.y + q.v.y*r.s + q.v.z*r.v.x - q.v.x*r.v.z;
	s.v.z = q.s*r.v.z + q.v.z*r.s + q.v.x*r.v.y - q.v.y*r.v.x;
	s.v.w = 1.0;

	return(s);
}

/* converts a quaternion to a rotation matrix */
void q_convert_m(quaternion q, GLdouble m[16])
{
	GLdouble xx = 0.0, xy = 0.0, xz = 0.0, xs = 0.0, yy = 0.0, yz = 0.0, ys = 0.0, zz = 0.0, zs = 0.0;

	xx = q.v.x*q.v.x;
	xy = q.v.x*q.v.y;
	xz = q.v.x*q.v.z;
	xs = q.v.x*q.s;
	yy = q.v.y*q.v.y;
	yz = q.v.y*q.v.z;
	ys = q.v.y*q.s;
	zz = q.v.z*q.v.z;
	zs = q.v.z*q.s;
/*
	m[0] = 1.0 - 2.0 * (yy + zz);
	m[1] = 2.0 * (xy - zs);
	m[2] = 2.0 * (xz + ys);
	m[4] = 2.0 * (xy + zs);
	m[5] = 1.0 - 2.0 * (xx + zz);
	m[6] = 2.0 * (yz - xs);
	m[8] = 2.0 * (xz - ys);
	m[9] = 2.0 * (yz + xs);
	m[10] = 1.0 - 2.0 * (xx + yy);
	m[3] = m[7] = m[11] = m[12] = m[13] = m[14] = 0.0;
	m[15] = 1.0;
*/

	m[0] = 1.0 - 2.0 * (yy + zz);
	m[4] = 2.0 * (xy - zs);
	m[8] = 2.0 * (xz + ys);
	m[1] = 2.0 * (xy + zs);
	m[5] = 1.0 - 2.0 * (xx + zz);
	m[9] = 2.0 * (yz - xs);
	m[2] = 2.0 * (xz - ys);
	m[6] = 2.0 * (yz + xs);
	m[10] = 1.0 - 2.0 * (xx + yy);
	m[3] = m[7] = m[11] = m[12] = m[13] = m[14] = 0.0;
	m[15] = 1.0;
}

/* convert rotation axis and angle to a quaternion */
quaternion aa_convert_q(vector axis, GLdouble angle)
{
	quaternion q;
	vector n;
	GLdouble sin_a = 0.0, cos_a = 0.0;

	n = normalize(axis);
	sin_a = sin((angle/2.0)*M_PI/180.0);
	cos_a = cos((angle/2.0)*M_PI/180.0);

	q.v.x = n.x * sin_a;
	q.v.y = n.y * sin_a;
	q.v.z = n.z * sin_a;
	q.v.w = 1.0;
	q.s = cos_a;

	return(q);
}

/* convert a quaternion to a rotation axis and an angle */
vector q_convert_aa(quaternion q, GLdouble *angle)
{
	vector axis;
	GLdouble sin_a = 0.0, cos_a = 0.0;

	q_normalize(q);
	cos_a = q.s;
	*angle = acos(cos_a) * 2.0;
	sin_a = sqrt(1.0 - cos_a * cos_a);
	if (fabs(sin_a) < 0.0005)
		sin_a = 1.0;
	
	axis.x = q.v.x/sin_a;
	axis.y = q.v.y/sin_a;
	axis.z = q.v.z/sin_a;
	axis.w = 1.0;

	return axis;
}

/* converts a vector into a quaternion */
quaternion v_convert_q(vector v)
{
	quaternion q;

	q.s = 0.0;
	q.v.x = v.x;
	q.v.y = v.y;
	q.v.z = v.z;
	q.v.w = 1.0;

	return(q);
}

/* applies a quaternion rotation to a vector through a matrix representation 
 * (changes the vector) */
void apply_q2v(quaternion q, vector *v)
{
	GLdouble m[16];

	q_convert_m(q, m);
	apply_v2m(v, m);
}
