/*------------------------------------------------------------------------------
* Alpha Shooter
* A 3D OpenGL First-Person Shooter game
* Copyright (C) 2005, 2006, 2007 Nicola Cocchiaro <ncocchiaro@users.sf.net>
* Alpha Shooter can be found at http://www.sf.net/projects/alphashooter.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 or, (at your option) any later
* version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*-------------------------------------------------------------------------------
*/

// texture.c
// some operations for texture mapping and inizialization

#include "texture.h"
#include "defs.h"
#include <errno.h>
#include <string.h>

/* sets defaults for textures (may be overridden later) */
void texture_defaults(GLint format)
{
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	gluBuild2DMipmaps(GL_TEXTURE_2D, format, tex->tex_data->sizeX, tex->tex_data->sizeY, format, GL_UNSIGNED_BYTE, tex->tex_data->data);

	free(tex->tex_data->data);
	tex->tex_data->data = NULL;
	free(tex->tex_data);
	tex->tex_data = NULL;
}

/* initializes a texture */
void init_texture(const char *filename, GLint tex_num, GLint format)
{
	GLuint tex_name = 0;
	int res = 0, i = 0, done = 0;
	char *effective_filename = NULL;
	size_t len = 0;

#if (!defined ALPSHOOT_WIN32)
	const char *alt_prefix[AS_NUM_ALT_PATHS] = { "/usr/share/games/alphashooter/", "/usr/local/share/games/alphashooter/", NULL };
#else
	const char *alt_prefix[AS_NUM_ALT_PATHS] = { NULL, NULL, NULL };
#endif
	
	// sanity check
	if ((!filename) || (tex_num < 0) || (format < 0))
		fprintf(stderr, "Error: Invalid parameters in init_texture: filename is %p, tex_num is %d, format is %d\n", filename, (int)tex_num, (int)format);

	len = strlen(filename)+1;
	
	effective_filename = (char*)malloc(len + 40);
	if (!effective_filename)
	{
		fprintf(stderr, "Error: Unable to allocate memory for effective_filename in init_texture: %s\n", strerror(errno));
		return;
	}
	memset(effective_filename, 0, len+40);
	
	strncat(effective_filename, filename, len);
	
	i = 0;
	// try to find texture in current directory, if that fails, try alternate paths
	while ((i <= AS_NUM_ALT_PATHS) && (!done))
	{
		tex->tex_data = loadTGA(effective_filename, &res);
		if ((tex->tex_data == NULL) || (tex->tex_data->data == NULL))
		{
			if (alt_prefix[i])
			{
				memset(effective_filename, 0, len+40);
				strncat(effective_filename, alt_prefix[i], strlen(alt_prefix[i]));
				strncat(effective_filename, filename, len);
				i++;
			}
			else
				// just quit with an error, there's nothing else to do
				i = AS_NUM_ALT_PATHS+1;
		}
		else
		{
			tex_name = tex->tex_names[tex_num];
			if ((tex_name == GL_INVALID_VALUE) || (tex_name == GL_INVALID_OPERATION) || (tex_name == 0))
			{
				fprintf(stderr, "Unable to generate texture name\n");
				exit(1);
			}
			glBindTexture(GL_TEXTURE_2D, tex->tex_names[tex_num]);
			texture_defaults(format);
			done = 1;
		}
	}
	if (i == AS_NUM_ALT_PATHS+1)
	{
		fprintf(stderr, "Unable to load texture %s, paths searched: %s", filename, "data/");
		for (i = 0; i < AS_NUM_ALT_PATHS; i++)
			if (alt_prefix[i])
				fprintf(stderr, ", %s", alt_prefix[i]);
		fprintf(stderr, ".\n");
	}
	free(effective_filename);
	return;
}
